/***************************************************************

	Windows version of Fishnet( Ported from Linux Version )
	Author: Rajesh Hegde K ( krhegde@cs.washington.edu )
	April 21, 2002

***************************************************************/

#include "winfishguts.h"
#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
//#include <errno.h>

typedef size_t ssize_t;
typedef unsigned __int32 in_addr_t;
#include <io.h>
#include <assert.h>
#include <math.h>
#include "getopt_l.h"
#include <crtdbg.h>

#define DEFAULT_LOSS_RATE (0.0)

/* standard nspring things */
/* not the greatest MAX ever, but it gets the point across */
#define MAX(a,b) ((a) < (b) ? (b) : (a))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

/* defaults and constants */
#define ACCEPT_QUEUE_LEN 10
static unsigned short serverport = DEFAULT_PORT;

/* fishnet node information */
typedef struct node 
{
  /*@null@*/
  struct node *next; /* it's a singly linked list! */

  /* for others to contact */
  int fishnet_address; 

  /* not pointers, for bug avoidance */
  int fishnet_neighbor_addresses[MAX_NEIGHBORS];
  
  /* stuff for management */
  int socket;
  char username[8];
  char domain[8];
  int gateway;

  /* to be distributed */
  struct in_addr ip_address; /* in network byte order */
  unsigned short udp_port;   /* in host byte order */
} node, *pnode;


/*@dependent@*/ /*@null@*/ 
static const node *find_by_address(int fishnet_address, /*@null@*/const node *nodelist);
static void broadcast_neighbor_list(/*@ null @*/ pnode nodelist);
static pnode make_new_node(int s_accepting, /*@null@*/ node *list);
static int needs_gateway(const char* domain, /*@null@*/ node *nodelist);
static int build_select_params(int accepting_socket, 
                                /*@ null @*/ const pnode nodelist,
                               /*@out@*/fd_set *pfdset);
/*@null@*/
static pnode remove_from_list_return_previous(/*@only@*/pnode p, pnode *plist);
static void send_error(pnode p, const char *message);
static void add_random_neighbors(pnode new_node, /*@null@*/ node *nodelist);
static void add_chain_neighbors(pnode new_node, /*@null@*/ node *nodelist);
static void add_clique_neighbors(pnode new_node, /*@null@*/ node *nodelist);
static void add_ring_neighbors(pnode new_node, /*@null@*/ node *nodelist);
static void add_hierarchical_neighbors(pnode new_node, /*@null@*/ node *nodelist);
static void clean_neighbor_list(/*@ null @*/ pnode nodelist);
static void send_web_response(int web_socket, /*@ null @*/ const pnode nodelist);

typedef struct 
{ 
  void (*add)(pnode newnode, node *list);
  const char *name;
} S_add_neighbours; 

S_add_neighbours add_neighbors_array[] = 
{
  { add_random_neighbors, "pseudo-random"},
  { add_chain_neighbors,  "chain"},
  { add_ring_neighbors,   "ring"},
  { add_clique_neighbors, "clique"},
  { add_hierarchical_neighbors, "hierarchy"},
  { NULL, NULL }
};

#define DEFAULT_TOPOLOGY "pseudo-random"
static const char *topology =  DEFAULT_TOPOLOGY;

void add_neighbor(pnode new_node, node *node_list) 
{
  int i;
  for(i=0; add_neighbors_array[i].add != NULL && stricmp(add_neighbors_array[i].name, topology); i++);
  
  if(add_neighbors_array[i].add != NULL) 
  {
    (add_neighbors_array[i].add)(new_node, node_list);
  } 
  else 
  {
    printf("unsupported topology: %s\n", topology);
    exit(EXIT_FAILURE);
  }
}


static void print_usage(void);
#define DEFAULT_LOSS_RATE (0.0)
float loss_rate = DEFAULT_LOSS_RATE;


int main(int argc, char * const argv[]) 
{
  int s_accepting;
  struct sockaddr_in addr_server;
  int r; /* return value */
  pnode nodelist = NULL;
  int c;
  char hostnamebuf[128];
  struct fish_message *pfmrecv = (struct fish_message *)malloc(1024);
  
  struct option longopts[] = 
  {
    { "port", required_argument, NULL, 'p' },
    { "debug", required_argument, NULL, 'd' },
    { "loss", required_argument, NULL, 'l' },
    { "help", no_argument, NULL, 'h' },
    { "extreme", no_argument, NULL, 'e' },
    { "topology", required_argument, NULL, 't' },
    { NULL, 0, NULL, 0 }
  };

  assert(pfmrecv!=NULL);

  memset(pfmrecv, 0, 1024);

  


  /* get server port number command line option */
  while((c = getopt_long(argc, argv, "p:d:l:t:he", longopts, NULL)) != -1) 
  { 
    switch(c) 
	{
    case 'p':

      serverport = (unsigned short int)atoi(optarg);
      if(serverport < 1024) 
	  {
        printf("WARNING: Ports less than 1024 are unlikely to work!\n");
      }
      break;

    case 'd':

      fish_setdebuglevel(atoi(optarg));
      break;

    case 'l':

      loss_rate = atof(optarg);
      if(loss_rate > 1.0 || loss_rate < 0.0) 
	  {
        printf("loss rate should be between 0 and 1, inclusive.\n");
        exit(EXIT_FAILURE);
      }
      printf("set loss rate to %f\n", loss_rate);
      break;

    case 'h':

      print_usage();
      exit(EXIT_SUCCESS);
      break;

    case 'e':

      fish_setdebuglevel(2);
      fish_setdebugfile(fopen("fishhead.log", "a"));
      break;

    case 't':

      topology = strdup(optarg);
      break;

    case ':':

      printf("ERROR: missing parameter on command line\n");
      print_usage();
      exit(EXIT_FAILURE);

    case '?':

      /* getopt() prints an ok error message... */
      print_usage();
      exit(EXIT_FAILURE);

    default:

      printf("ERROR: unexpected error on command line\n");
      print_usage();
      exit(EXIT_FAILURE);

    }
  }

  WORD wVersionRequested;
  WSADATA wsaData;
  int err;

  wVersionRequested = MAKEWORD( 2, 2 );

  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) 
  {
	/* Tell the user that we could not find a usable */
	/* WinSock DLL.                                  */
	return -1;
  }

  /* create server socket */
  s_accepting = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(s_accepting < 0) 
  {
    perror("socket() failed");
	WSACleanup();
    exit(EXIT_FAILURE);
  }

  memset(&addr_server, 0, sizeof(struct sockaddr_in));
  addr_server.sin_family = AF_INET;
  addr_server.sin_addr.s_addr = htonl(INADDR_ANY);
  addr_server.sin_port = htons(serverport);
  
  /* bind server socket */
  r = bind(		s_accepting,
				(struct sockaddr *) &addr_server,
				(socklen_t)sizeof(struct sockaddr_in));
  if(r != 0) 
  {
    perror("ERROR: bind() failed");
    printf("INFO: maybe supply a port other than %d with -p or wait a while?\n", serverport);
	WSACleanup();
    exit(EXIT_FAILURE);
  }

  /* listen on server socket */
  r = listen(s_accepting, ACCEPT_QUEUE_LEN);
  if(r < 0) 
  {
    perror("ERROR: listen() failed");
	WSACleanup();
    exit(EXIT_FAILURE);
  }
  
  /* could get domain name too, but that's less fun. */
  gethostname(hostnamebuf, 128);
  fish_debug(2, "fishhead started successfully on %s:%d\n", hostnamebuf, serverport);
  printf("Terminate fishhead when done with ^C.\n");

  while(TRUE) 
  {
	  
    /* flag that tells us whether to broadcast topology changes at the
       end of a select() run */
    boolean pending_update = FALSE;

    fd_set readfds;  /* bitmask of select sockets */
	memset( &readfds, 0, sizeof( fd_set ) );
    pnode p;         /* list iterator */
    int max_fd = build_select_params(s_accepting, nodelist, &readfds);
    int num_fds;     /* sockets returned by select; may be zero
                        if timer expired */

    /*** select sockets ***/

    /*@-nullpass@*/
    num_fds = select(max_fd+1, &readfds, NULL, NULL, NULL);
    /*@+nullpass@*/
    /* printf("select returned %d sockets\n", num_fds); */

   if(FD_ISSET(s_accepting, &readfds) != 0) 
   {
      /* accept a connection by creating a new node */
      nodelist = make_new_node(s_accepting, nodelist);
   }
    
    if(nodelist !=NULL) 
      for(p=nodelist; p!=NULL; p=p->next) 
	  {
      NEXT:
        if(FD_ISSET(p->socket, &readfds)!=0) 
		{
	
          /* read message */
          ssize_t readlen;
          readlen = recv( p->socket, (char*)pfmrecv, 1024, 0);

          if( readlen == SOCKET_ERROR || readlen == 0) 
		  {
            /* socket closed or read otherwise failed */

            pnode next = p->next;
            
            fish_debug(FISH_DEBUG_ALL, "ERROR: read() failed: %d\n", WSAGetLastError() );
           
            fish_debug(2, "leave: %s from %s\n", p->username, inet_ntoa(p->ip_address));
	  
            /* somewhat inelegant removal of a node then
               resumption of for loop */
            fish_debug(FISH_DEBUG_TOPOLOGY,"INFO: Removing %d (%s)\n", p->fishnet_address, inet_ntoa(p->ip_address));
            (void)remove_from_list_return_previous(p, &nodelist);
            close(p->socket);
            pending_update = TRUE;
            p=next;
            if(p) goto NEXT;
            else goto DONE;
	  
          } 
		  else if(readlen > 5 && (strncmp((const char*)pfmrecv, "GET ", 4) == 0)) 
		  {
            pnode next = p->next;
          
            fish_debug(2, "served topology info by web to %s\n", inet_ntoa(p->ip_address));

            /* web request */
            send_web_response(p->socket, nodelist);

            (void)remove_from_list_return_previous(p, &nodelist);
            p=next;
            if(p) goto NEXT;
            else goto DONE;

          } 
		  else if(pfmrecv->length < MIN_MESSAGE_SIZE || pfmrecv->length > 1024) 
		  {
            fish_debug(FISH_DEBUG_NONE,"WARNING: Received badly formed message\n");
            fprintf(stderr, "%s\n", (const char *)pfmrecv);
          } 
		  else if((ssize_t)pfmrecv->length > readlen) 
		  {
            /* TODO: handle this by buffering */
            fish_debug(FISH_DEBUG_NONE,"WARNING: Read too little data (%u) from message that's %u long\n",(unsigned int)readlen, pfmrecv->length);

          } 
		  else 
		  {
            if((ssize_t)pfmrecv->length < readlen) 
			{
              /* TODO: find out if this happens, if so, fix. */
              fish_debug(FISH_DEBUG_NONE,"WARNING: May have received two messages, dumping second\n");
            }

            fish_debug(FISH_DEBUG_ALL,"INFO: Message from %s:\n", inet_ntoa(p->ip_address));
            if (fish_getdebuglevel() >= FISH_DEBUG_ALL) 
			{
              printmessage(pfmrecv);
            }

            switch(pfmrecv->type) 
			{
            case FISH_MSG_TYPE_ERROR:
              break;
            case FISH_MSG_TYPE_NEIGHBORLIST:
              break;
            case FISH_MSG_TYPE_JOINREQUEST:
              /* ensure that address is not already taken */
              if(find_by_address(pfmrecv->body.joinreq.node_address, nodelist)) 
			  {
                send_error(p, "Address already in use");
              } 
			  else 
			  {
                /* process join request */
                p->fishnet_address = pfmrecv->body.joinreq.node_address;
                p->udp_port        = pfmrecv->body.joinreq.node_udp_port;
                strncpy(p->username, pfmrecv->body.joinreq.username, 8);
                strncpy(p->domain, pfmrecv->body.joinreq.domain, 8);

				p->gateway = FALSE;
				p->gateway = needs_gateway(p->domain, nodelist);

                fish_debug(FISH_DEBUG_TOPOLOGY,"INFO: Adding %s:%u to network (%s:%u,%8s)\n",
							p->domain,
							p->fishnet_address, 
							inet_ntoa(p->ip_address),
							p->udp_port, 
							p->username);
                add_neighbor(p, nodelist);
                fish_debug(2, "join: %s from %s:%d addr %d\n", 
                           p->username, 
                           inet_ntoa(p->ip_address), 
                           p->udp_port,
                           p->fishnet_address);
                pending_update = TRUE;
              }
              break;
            default:
              break;
            } /* end switch */
          } /* end message */
        } /* end if set */
      } /* end for loop */
  DONE:
    if(pending_update) 
	{
      clean_neighbor_list(nodelist);
      broadcast_neighbor_list(nodelist);
    }
  }
  WSACleanup();
}

/* build select parameters from node list */
static int build_select_params(int accepting_socket, 
                               /*@null@*/ const pnode nodelist,
                               /*@out@*/fd_set *pfdset) 
{
  pnode p;
  int max_fd = accepting_socket;

  FD_ZERO(pfdset);
  FD_SET( accepting_socket, pfdset );

  for(p=nodelist; p!=NULL; p=p->next) 
  {
    FD_SET( p->socket, pfdset );
    max_fd = MAX(max_fd, p->socket);
  }

  return(max_fd);
}

/*** message generation ***/
static void broadcast_neighbor_list(/*@null@*/ pnode nodelist) 
{
  pnode p;
  struct fish_message *pfm = (struct fish_message *)malloc(1024);

  fish_debug(FISH_DEBUG_TOPOLOGY, "INFO: Broadcasting neighbor list\n");

  if(pfm) 
  {
    memset(pfm, 0, 1024);
    pfm->type = FISH_MSG_TYPE_NEIGHBORLIST;
    
    /* for each node */
    for(p=nodelist; p!=NULL; p=p->next) 
	{
      int i,j;
      char nodemsg[80];
      char *m = nodemsg;
      m+=sprintf(nodemsg,/* 80,*/ "  %d: ", p->fishnet_address);

      /* construct array of neighbor structures given addresses */
      for(i=0,j=0; i< MAX_NEIGHBORS; i++) 
	  {
        int neighbor_address = p->fishnet_neighbor_addresses[i];
        if(neighbor_address != 0) 
		{
          const node *n = find_by_address(neighbor_address, nodelist);
          if(n) 
		  {
            m+=sprintf(m, /*80-(m-nodemsg),*/ "%d ", neighbor_address);
            pfm->body.neighbors[j].address         = neighbor_address;
            pfm->body.neighbors[j].node_ip_address = n->ip_address;
            pfm->body.neighbors[j].node_udp_port   = n->udp_port;
            pfm->body.neighbors[j].lossrate_x_1000 =  MIN(1000U, (unsigned short)(loss_rate * (float)1000)), j++;
          } 
		  else 
		  {
            fish_debug(3, "Error in neighbors list, zeroing for later consistency\n");
            p->fishnet_neighbor_addresses[i] = 0;
          }
        }
      }
      fish_debug(FISH_DEBUG_TOPOLOGY, "%s\n", nodemsg);
      pfm->length = MIN_MESSAGE_SIZE + (sizeof(struct neighbor) * j);

      /* send neighbor message for this node */
      int byteswritten = send(p->socket, (char*)pfm, (size_t)pfm->length, 0);
	  if( byteswritten < 0 )
	  {
		  assert( FALSE );
	  }


    }

    free(pfm);
    
  } else {
    printf("ERROR: unable to allocate packet to broadcast neighbor list.\n");
    exit(EXIT_FAILURE);
  }
}

/* send error to fishnet node */
static void send_error(pnode p, const char *message) 
{
  unsigned short length = MIN_MESSAGE_SIZE + strlen(message) + 1;
  struct fish_message *pfm = (fish_message*)malloc((size_t)length);
  
  if(pfm) 
  {
    pfm->length = length;
    pfm->type = FISH_MSG_TYPE_ERROR;
    strcpy(pfm->body.errormessage, message);
    (void)send(p->socket, (char*)pfm, (size_t)pfm->length, 0);
    free(pfm);
  }/* else, we're too fsck'd up to care. */
}

/* decide whether a particular domain needs a gateway */
static int needs_gateway(const char* domain, node* nodelist) 
{
  node* node_n;
  for(node_n=nodelist; node_n!=NULL; node_n=node_n->next) 
  {
    if(node_n->fishnet_address && (0 == strcmp(domain, node_n->domain)) && node_n->gateway) 
	{
      // This domain already has a gateway
      return FALSE; 
    }
  }
  return TRUE;
}

/* find a node by its fishnet address */
/*@dependent@*//*@null@*/ 
static const node * find_by_address(int fishnet_address, 
                             /*@null@*/const node *nodelist) 
{
  const node *p;

  for(p=nodelist; (p!=NULL) && (p->fishnet_address != fishnet_address);  p=p->next);

  return(p);
}

/*** node management: creation ***/
/*@only@*/ static pnode make_new_node(int s_accepting, /*@null@*/node *list) 
{
  struct sockaddr_in addr_client;
  socklen_t len_addr_client = (socklen_t)sizeof(struct sockaddr_in);
  pnode new_node;
  
  int newsocket = accept(	s_accepting, 
							(struct sockaddr *) &addr_client,
							&len_addr_client);
  if(newsocket < 0) 
  {
    fish_debug(FISH_DEBUG_INTERNAL,"WARNING: select told us to accept, ""but it didn't work: %s", strerror(GetLastError()));
    return (NULL);
  }

  fish_debug(FISH_DEBUG_TOPOLOGY,"INFO: Connection from %s\n", inet_ntoa(addr_client.sin_addr));

  new_node = ( pnode )malloc(sizeof(node));
  assert(new_node != NULL);
  memset(new_node, 0, sizeof(node));
  new_node -> socket = newsocket;
  new_node -> ip_address = addr_client.sin_addr;
  new_node -> next = list;
  /* prepend to list is done by caller */

  return(new_node);
}

/*** node management: removal ***/
/* TODO: cleanup neighbors list? */
/*@null@*/
static pnode remove_from_list_return_previous(/*@only@*/ pnode p, 
                                              pnode *plist) 
{
  pnode q;
  assert(p!=NULL);
  if(p==*plist) 
  { 
    /* printf("removing head of list\n"); */
    *plist = p->next;
    free(p);
    return(NULL);
  }
  for(q = *plist; q != NULL && q->next !=p; q=q->next);

  if(q != NULL) 
  {
    /* printf("found the node to remove\n"); */
    q->next = q->next->next;
    free(p);
  } 
  else 
  {
    /* if we didn't find it's predecessor, and it's not first...  */
    fish_debug(FISH_DEBUG_ALL, "WARNING: strange removal\n");
    free(p);
  }
  return q;
}

/* neighbor list management: query */
static boolean is_in_neighbor_list(const node *haystack_node, 
                                   int needle_address) 
{
  int i;
  for(i=0; i<MAX_NEIGHBORS; i++) 
  {
    if(haystack_node->fishnet_neighbor_addresses[i] == needle_address) 
      return TRUE;
  }
  return FALSE;
}

#ifdef unused
static boolean is_a_neighbor_by_node(const pnode node1, const pnode node2) 
{
  return(node1 !=NULL && node2 !=NULL &&
         is_in_neighbor_list(node1, node2->fishnet_address) &&
         is_in_neighbor_list(node2, node1->fishnet_address));
}
#endif

/* neighbor list management: addition
 * (observes the invariant that no node can have more than MAXNEIGHBORS
 * neighbors) */
static boolean add_to_neighbor_list(node *n, int neighbor_address) 
{
  int i;
  for(i=0; i<MAX_NEIGHBORS &&  n->fishnet_neighbor_addresses[i] != 0 ; i++);

  if(i<MAX_NEIGHBORS) 
  {
    n->fishnet_neighbor_addresses[i] = neighbor_address;
    return (TRUE);
  } 
  else 
  {
    fish_debug(FISH_DEBUG_INTERNAL, "ERROR: node %d has too many neighbors", n->fishnet_address);
    return (FALSE);
  }
}

static boolean remove_from_neighbor_list(pnode n, int neighbor_address) 
{
  int i;
  for(i=0; i<MAX_NEIGHBORS && n->fishnet_neighbor_addresses[i] != neighbor_address ; i++);

  if(i<MAX_NEIGHBORS) 
  {
    n->fishnet_neighbor_addresses[i] = 0;
    return (TRUE);
  } 
  else 
  {
    fish_debug(FISH_DEBUG_INTERNAL, "WARN: didn't find %d in %d's neighbor list for removal", neighbor_address, n->fishnet_address);

    return (FALSE);
  }
}
/* how many neighbors */
static unsigned int count_nodes_neighbors(const pnode n) 
{
  unsigned int i;
  for(i=0; i<MAX_NEIGHBORS &&  n->fishnet_neighbor_addresses[i] != 0 ; i++);

  return(i);
}


/* neighbor list management: cleanup */
static void clean_neighbor_list(/*@ null @*/ pnode nodelist) 
{
  pnode p;
  fish_debug(FISH_DEBUG_INTERNAL, "INFO: Cleaning neighbor list\n");
  /* foreach node */
  for(p=nodelist; p!=NULL; p=p->next) 
  {
    int i;
    /* foreach neighbor */
    for(i=0; i< MAX_NEIGHBORS; i++) 
	{
      int neighbor_address = p->fishnet_neighbor_addresses[i];
      if(neighbor_address!=0) 
	  {
        const node *n = find_by_address(neighbor_address, nodelist);
        if(n) 
		{
          if(!is_in_neighbor_list(n,p->fishnet_address)) 
		  {
            (void)add_to_neighbor_list((node *)n, p->fishnet_address);
            fish_debug(FISH_DEBUG_INTERNAL, "INFO: Mending neighbor list\n");
          }
        } 
		else 
		{
          fish_debug(FISH_DEBUG_INTERNAL, "INFO: Node vanished; zeroing\n");
          p->fishnet_neighbor_addresses[i] = 0;
        }
      }
    }
  }
}

typedef int approval_function(node*, node*);

/* gets the number of nodes in the list that are OK to add */
static unsigned int num_nodes(/*@null@*/const node *nlist,
		              node* new_node, 
	                      approval_function ok_to_add) 
{
  const node *q;
  unsigned int numnodes = 0;
  for(q = nlist; q != NULL; q=q->next) 
  {
    if (ok_to_add((node*)new_node, (node*)q)) 
	{
      numnodes++;
    }
  }
  return numnodes; 
}

static boolean try_breaking_adjacency(pnode a, pnode b) 
{
  return( remove_from_neighbor_list(a, b->fishnet_address) &&
          remove_from_neighbor_list(b, a->fishnet_address));
  
}

static boolean try_creating_adjacency(pnode a, pnode b) 
{
  in_addr_t localhost = inet_addr("127.0.0.1");
 
  if((a->ip_address.s_addr == localhost
      && b->ip_address.s_addr == localhost)
     || (a->ip_address.s_addr != localhost
         && b->ip_address.s_addr != localhost)) 
  {
    /* attempt to add to other's neighbor list */
    if(add_to_neighbor_list(a, b->fishnet_address)) 
	{
      /* only complete our add if the other one succeeded */
      (void)add_to_neighbor_list(b, a->fishnet_address);

      fish_debug(FISH_DEBUG_TOPOLOGY, "INFO: paired %d with %d\n", b->fishnet_address, a->fishnet_address);

      return(TRUE);
    }
  } 
  else 
  {
    fish_debug(FISH_DEBUG_NONE,
               "WARNING: NOT CONNECTING new node %d with %d:\n"
               "One has connected to fishead using localhost\n"
               "The other using fishhead's hostname\n",
               b->fishnet_address,
               a->fishnet_address);
  }
  return(FALSE);
}

static void add_clique_neighbors(pnode new_node, 
                                 /*@null@*/node *nodelist) 
{
  pnode list_iterator;

  /* find a different node with at most one neighbor */
  for(list_iterator = nodelist;  list_iterator!=NULL && count_nodes_neighbors(new_node) < MAX_NEIGHBORS; list_iterator=list_iterator->next) 
  {
    if(list_iterator != new_node) 
	{
      if(count_nodes_neighbors(list_iterator) < MAX_NEIGHBORS) 
	  { 
        try_creating_adjacency(list_iterator, new_node);
      }
    }
  }
}

/* Add neighbor into the chain.  Verify that it's a chain. */
/* TODO: reestablish chain if an intermediate link is removed
   and replaced */
static void add_chain_neighbors(pnode new_node, /*@null@*/node *nodelist) 
{
  pnode list_iterator;

  /* find a different node with at most one neighbor */
  for(list_iterator = nodelist; list_iterator!=NULL; list_iterator=list_iterator->next) 
  {
    if(list_iterator != new_node && count_nodes_neighbors(list_iterator) <= 1) 
	{ /* isn't in the middle */
      if(try_creating_adjacency(list_iterator, new_node)) 
	  {
        return; /* done */
      }
    }
  }
}
  
/* Add neighbor into the ring.  Verify that it's a chain. */
static void add_ring_neighbors(pnode new_node, /*@null@*/node *nodelist) 
{
  pnode list_iterator;

  int our_degree = 0;
  
  pnode left = NULL;
  
  /* first, see if there are two nodes with degree 1, use those
     to merge the ring. */
  for(list_iterator = nodelist; list_iterator!=NULL && our_degree < 2; list_iterator=list_iterator->next) 
  {
    if(list_iterator != new_node) 
	{
      /* collect orphans and stubs */
      if(count_nodes_neighbors(list_iterator) <= 1) 
	  { 
        try_creating_adjacency(new_node, list_iterator);
        our_degree ++;
      }
    }
  }
  
  /* we've done our part; if we have one neighbor, we can't 
     very well splice into a ring.. */
  if(our_degree >= 1) 
  {
    return;
  }
  
  /* can get here if we're the only node or there is only one other, 
     so find two that aren't our own before splicing */
  
  for(list_iterator = nodelist; list_iterator!=NULL; list_iterator=list_iterator->next) 
  {
    if(list_iterator != new_node) 
	{
      if(left == NULL) 
	  {
        left = list_iterator;
      } 
	  else 
	  {
        try_breaking_adjacency(left, list_iterator);
        try_creating_adjacency(left, new_node);
        try_creating_adjacency(list_iterator, new_node);
        return;
      }
    }
  }
}

/* Add random neighbors to a new node. */
static void add_random_neighbors_util(pnode new_node, 
		                      /*@null@*/node *nodelist,
				      approval_function ok_to_add) 
{
  unsigned int numnodes = num_nodes(nodelist, new_node, ok_to_add);
  unsigned int choose =  MIN(numnodes-1, 1 + (int)floor(log(numnodes)));
  unsigned int nodes_seen;
  pnode list_iterator;

  fish_debug(FISH_DEBUG_INTERNAL, "INFO: add %u of %u neighbors\n", choose, numnodes - 1);
  if(choose == 0) return; /* no neighbors to add */

  for(nodes_seen=0, list_iterator = nodelist; list_iterator!=NULL; list_iterator=list_iterator->next) 
  {
    if((list_iterator != new_node) && ok_to_add(new_node, list_iterator)) 
	{
      unsigned int rnd = (unsigned int)rand();
      /* probability of choosing a node is thresh / RAND_MAX */
      /* which is choose / numnodes - nodes_seen*/
      /* which is the number of additional neighbors we want
         divided by the number of nodes remaining in the list
         (other than us (which is why i starts at 1) */
      /* if rand is less than thresh, add node */
      unsigned int thresh = MIN((unsigned int)RAND_MAX, 
                                (unsigned int)((float)RAND_MAX * (float) choose / 
                                               (float)(numnodes - 1 - nodes_seen)));
      nodes_seen++;
      
      fish_debug(FISH_DEBUG_INTERNAL, "INFO: %u <=? %u\n", rnd, thresh);
	         	 
      if(rnd <= thresh) 
	  {
        if(try_creating_adjacency(list_iterator, new_node)) 
		{
          choose--;
        }
      }
    }
  }
}

int ok_to_add_random(node* new_node, node* other_node) 
{
  return TRUE;
}

/* Add random neighbors to a new node.  Observes the invariant that no
 * node can have more than MAX_NEIGHBORS neighbors. */
static void add_random_neighbors(pnode new_node, 
		                 /*@null@*/node *nodelist) 
{
  add_random_neighbors_util(new_node, nodelist, ok_to_add_random);

}
  
static int getsubnet(int address) 
{
  return (address < 1000) ? 0 : (address/10);
}

int ok_to_add_subnet(node* new_node, node* other_node) 
{
  return (getsubnet(new_node->fishnet_address) ==
		  getsubnet(other_node->fishnet_address));
}

int ok_to_add_internet(node* new_node, node* other_node) 
{
  return (other_node->fishnet_address < 1000 
	  || other_node->fishnet_address % 10 == 0);
}
  
/* Add neighbors to a new node to form a hierarchical network
 * according to subnets. */
static void add_hierarchical_neighbors(pnode new_node, 
		                       /*@null@*/node *nodelist) 
{
	 
  if (new_node->fishnet_address >= 1000) 
  {
    fish_debug(FISH_DEBUG_INTERNAL, "Adding subnet neighbors\n");
    add_random_neighbors_util(new_node, nodelist, ok_to_add_subnet);
  }

  if (new_node->fishnet_address < 1000 || (new_node->fishnet_address % 10 == 0)) 
  {
    fish_debug(FISH_DEBUG_INTERNAL, "Adding internet neighbors\n");
    add_random_neighbors_util(new_node, nodelist, ok_to_add_internet);
  }
}
  
          
static void send_web_response(int web_socket, const pnode nodelist) 
{

#define WEB_RESP_LEN 8192
  char response[WEB_RESP_LEN]; 
  int resp_len=0;
  node const *node_n;
  int e;
  int nodecount = 0;
  resp_len = sprintf(response,/* WEB_RESP_LEN,*/ "HTTP/1.0 200 OK\n"
                      "Content-Type: text/html\n"
                      "\n"
                      "<html><head><title>Fishnet Topology</title></head>\n"
                      "<body>\n"
                      "<h2>FishWeb Servers</h2>\n");
  for(node_n=nodelist; node_n!=NULL; node_n=node_n->next) 
  {
    if(node_n->fishnet_address) 
	{
      resp_len += sprintf(&response[resp_len],/* WEB_RESP_LEN-resp_len, */
                           " <a href=\"http://%u/\">%u</a> (%s)<br>\n",
                           node_n->fishnet_address, 
                           node_n->fishnet_address,
                           node_n->username);
      nodecount ++;
    }
  }
  
  resp_len += sprintf(&response[resp_len],/* WEB_RESP_LEN-resp_len, */
                       "<h2>Fishnet Topology</h2>\n"
                       "<img src=\"http://poplar.cs.washington.edu/cgi-bin/dotter.pl?");
  e=1;
  
  for(node_n=nodelist/*, n=1*/; node_n!=NULL; node_n=node_n->next/*, n++*/) 
  {
    if(node_n->fishnet_address) 
	{
      int i;
      resp_len += sprintf(&response[resp_len], /*WEB_RESP_LEN-resp_len, */
							"n%d=%s_%s&", 
							node_n->fishnet_address,
							node_n->username,
							node_n->domain);

      for(i=0; i< MAX_NEIGHBORS; i++) 
	  {
        int neighbor_address = node_n->fishnet_neighbor_addresses[i];
        /* get unidirectional lines (cleaner) if the graph is 
           at least 4 nodes large */
        if(neighbor_address != 0 && (nodecount < 4 || node_n->fishnet_address < neighbor_address )) 
		{
          resp_len+=sprintf(&response[resp_len], /*WEB_RESP_LEN-resp_len,*/
			                     "e%d=%d,%d&",
				                 e++,
					             node_n->fishnet_address, 
						         neighbor_address);
        }          
      }
    }
  }
  resp_len--; /* gobble trailing &. */
  resp_len+=sprintf(&response[resp_len], 
                     /*WEB_RESP_LEN-resp_len, */
                     "\">");
  resp_len+=sprintf(&response[resp_len], 
                     /*WEB_RESP_LEN-resp_len, */
                     "</body></html>");
  (void)send(web_socket, (char*)response, resp_len, 0);
  close(web_socket);
}
  
/* Prints out the usage message*/
static void print_usage(void) 
{
  printf("Usage: fishhead [options]\n"
         "Run the Fishnet network manager.\n"
         "\n"
         "options can include the following, where single character\n"
         "options are a shorthand for the longer ones..\n"
         "\n"
         " -h, --help                print this message\n"
         " -p, --port <port>         run fishhead on the specified port\n"
         "                           (default: %d)\n"
         " -d, --debug <debuglevel>  use the specified debug level\n"
         "                           (default: 7)\n"
         " -l, --loss <loss rate>    add loss to all links.\n"
         "                           (default: %f)\n"
         " -t, --topology <topology> use a specified topology:\n"
         "                           chain, clique, ring, pseudo-random, hierarchy\n"
         "                           (default: %s)\n"
         "\n"
         "Examples: ./fishhead -p 2256 -l 0.1\n"
         "          ./fishhead --port 2256  --topology chain\n"
         "          ./fishhead --port 2256 --loss 0.1 --topology chain\n"
         "\n",
         DEFAULT_PORT,
         DEFAULT_LOSS_RATE,
         DEFAULT_TOPOLOGY);
  /* exit is always called after print_usage */
}
